"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const crypto = require("crypto");
const atma_io_1 = require("atma-io");
const zlib = require("zlib");
const CACHE_BASE = './cache/squery';
class Cache {
    constructor() {
        this.meta = null;
        this.isFlushDeferred = false;
    }
    hasInner(url, config, { isAsync = false } = {}) {
        if (config.cache == null || config.cache === false) {
            return false;
        }
        url = this.normalizeUrl(url, config);
        let domainKey = Utils.getDomainKey(url);
        this.ensureMeta(domainKey);
        let meta = this.meta[domainKey][url];
        if (meta == null) {
            return false;
        }
        return atma_io_1.File[isAsync ? 'exists' : 'existsAsync'](`${CACHE_BASE}/${domainKey}/${meta.file}`);
    }
    has(url, config) {
        return this.hasInner(url, config, { isAsync: false });
    }
    hasAsync(url, config) {
        return this.hasInner(url, config, { isAsync: true });
    }
    remove(url, config) {
        url = this.normalizeUrl(url, config);
        let domainKey = Utils.getDomainKey(url);
        this.ensureMeta(domainKey);
        let meta = this.meta[domainKey][url];
        if (meta == null) {
            return null;
        }
        delete this.meta[domainKey][url];
        this.flushMeta(domainKey);
    }
    get(url, config) {
        return __awaiter(this, void 0, void 0, function* () {
            if (config.cache == null || config.cache === false) {
                return null;
            }
            url = this.normalizeUrl(url, config);
            let domainKey = Utils.getDomainKey(url);
            this.ensureMeta(domainKey);
            let meta = this.meta[domainKey][url];
            if (meta == null) {
                return null;
            }
            let now = Date.now();
            let seconds = ((now - meta.time) / 1000) | 0;
            if (meta.maxAge && seconds > meta.maxAge) {
                return null;
            }
            let withCompression = meta.file.endsWith('.gz');
            let encoding = withCompression ? 'buffer' : 'utf8';
            let result = yield new atma_io_1.File(`${CACHE_BASE}/${domainKey}/${meta.file}`, { cached: false }).readAsync({ encoding });
            if (withCompression) {
                let str = yield Compression.decompress(result);
                result = JSON.parse(str);
            }
            if (result.file != null) {
                result.body = yield new atma_io_1.File(`${CACHE_BASE}/${domainKey}/${result.file}`, { cached: false }).readAsync({ encoding: 'buffer' });
                delete result.file;
            }
            return result;
        });
    }
    save(url, config, resp) {
        if (config.cache == null || config.cache === false) {
            return null;
        }
        let cache = typeof config.cache !== 'boolean' ? config.cache : {
            compress: true,
            maxAge: 24 * 60 * 60
        };
        if (config.cache === true) {
            cache = {
                compress: true,
            };
        }
        url = this.normalizeUrl(url, config);
        let domainKey = Utils.getDomainKey(url);
        this.ensureMeta(domainKey);
        let md5 = crypto.createHash('md5').update(url).digest('hex');
        let file = `${md5}.json`;
        let withCompression = cache.compress;
        if (withCompression) {
            file += '.gz';
        }
        this.meta[domainKey][url] = {
            time: Date.now(),
            file: file,
            maxAge: cache.maxAge
        };
        this.flushMeta(domainKey);
        let json = {
            status: resp.status,
            headers: resp.headers,
            url: resp.url,
            body: resp.body
        };
        let contentType = resp.headers['content-type'];
        let isText = /json|text/.test(contentType);
        if (isText === false) {
            let match = /\.[\w\d]+$/.exec(url);
            let ext = match[0];
            let path = `${CACHE_BASE}/${domainKey}/files/${md5}${ext}`;
            let file = new atma_io_1.File(path, { cached: false });
            file.writeAsync(resp.body);
            json.file = `files/${md5}${ext}`;
            delete json.body;
        }
        if (!withCompression) {
            new atma_io_1.File(`${CACHE_BASE}/${domainKey}/${file}`, { cached: false }).writeAsync(json);
            return;
        }
        let str = JSON.stringify(json);
        Compression.compress(Buffer.from(str)).then(buffer => {
            new atma_io_1.File(`${CACHE_BASE}/${domainKey}/${file}`, { cached: false }).writeAsync(buffer);
        });
    }
    normalizeUrl(url, config) {
        url = url.toLowerCase().replace(/(?<!:)[/]{2,}/g, '/');
        let ignore = config.cacheQueryIgnore;
        if (ignore) {
            ignore.forEach(x => {
                url = url.replace(new RegExp(`&${x}=[\\w\\d]+`), '');
                url = url.replace(new RegExp(`\\?${x}=[\\w\\d]+`), '?');
            });
        }
        return url;
    }
    ensureMeta(domainKey) {
        if (this.meta != null && this.meta[domainKey] != null) {
            return;
        }
        if (this.meta == null) {
            this.meta = {};
        }
        let file = `${CACHE_BASE}/${domainKey}/meta.json`;
        if (atma_io_1.File.exists(file)) {
            this.meta[domainKey] = atma_io_1.File.read(file);
        }
        else {
            this.meta[domainKey] = {};
        }
    }
    flushMeta(domainKey) {
        if (this.isFlushDeferred) {
            return;
        }
        this.isFlushDeferred = true;
        setTimeout(() => {
            atma_io_1.File
                .writeAsync(`${CACHE_BASE}/${domainKey}/meta.json`, this.meta[domainKey])
                .always(x => {
                this.isFlushDeferred = false;
            });
        }, 1000);
    }
}
exports.Cache = Cache;
class Compression {
    static compress(buffer) {
        return new Promise((resolve, reject) => {
            zlib.deflate(buffer, (err, buffer) => {
                if (err) {
                    reject(err);
                    return;
                }
                resolve(buffer);
            });
        });
    }
    static decompress(buffer) {
        return new Promise((resolve, reject) => {
            zlib.unzip(buffer, (err, buffer) => {
                if (err) {
                    reject(err);
                    return;
                }
                resolve(buffer.toString('utf8'));
            });
        });
    }
}
var Utils;
(function (Utils) {
    function getDomainKey(url) {
        let domainMatch = /([\w\d_\-]+\.[\w]{2,5})([\/:]|$)/.exec(url);
        if (domainMatch) {
            return domainMatch[1].replace('.', '_');
        }
        return '';
    }
    Utils.getDomainKey = getDomainKey;
})(Utils || (Utils = {}));
exports.cache = new Cache;
