"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const fetch = require("node-fetch");
const https = require("https");
const http = require("http");
const CookieContainer_1 = require("../common/CookieContainer");
const Cache_1 = require("./Cache");
const atma_utils_1 = require("atma-utils");
const Body_1 = require("./Body");
const DefaultOptions = {
    headers: {
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8',
        'Accept-Encoding': 'gzip, deflate',
        'Accept-Language': 'en,ru;q=0.9,de;q=0.8,en-GB;q=0.7,uk;q=0.6,la;q=0.5',
        'Cache-Control': 'no-cache',
        'Pragma': 'no-cache',
        'Referer': 'https://www.google.de/',
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Safari/537.36',
    }
};
const agents = {
    http: new http.Agent({ keepAlive: true }),
    https: new https.Agent({ keepAlive: true }),
};
exports.NetworkDriver = {
    isCached(url, config = {}) {
        url = serializeUrl(url, config);
        return Cache_1.cache.has(url, config);
    },
    isCachedAsync(url, config = {}) {
        url = serializeUrl(url, config);
        return Cache_1.cache.hasAsync(url, config);
    },
    clearCookies() {
        CookieContainer_1.cookieContainer.clearCookies();
    },
    clearCached(url, config = {}) {
        url = serializeUrl(url, config);
        Cache_1.cache.remove(url, config);
    },
    load(url, config = {}) {
        let worker = new RequestWorker(url, config);
        return worker.load();
    },
    getCookies(url) {
        return CookieContainer_1.cookieContainer.getCookies(url);
    },
    setCookies: ((...args) => {
        CookieContainer_1.cookieContainer.addCookies.apply(CookieContainer_1.cookieContainer, args);
    })
};
function serializeUrl(url, config = {}) {
    if (config.query) {
        let q = '';
        for (let key in config.query) {
            let p = `${key}=${encodeURIComponent(config.query[key])}`;
            q += (q ? '&' : '') + p;
        }
        url += (url.includes('?') ? '&' : '?') + q;
    }
    return url;
}
function readAllHeaders(headers) {
    let obj = {};
    for (let entry of headers.entries()) {
        let [key, value] = entry;
        obj[key] = value;
    }
    return obj;
}
class RequestWorker {
    constructor(url, config = {}) {
        this.url = url;
        this.config = config;
        this.redirectIndex = 0;
        this.retryIndex = 0;
        this.options = {
            headers: Object.assign({}, DefaultOptions.headers, config.headers || {}),
            method: config.method,
            body: config.body,
            follow: config.follow,
            onRedirect(data) {
                if (data.prev.startsWith('http:') && data.url.includes('https:')) {
                    data.opts.agent = agents.https;
                }
            }
        };
        this.cookieContainer = config.cookieContainer || CookieContainer_1.cookieContainer;
        this.retryCount = 'retryCount' in config ? config.retryCount : 3;
        this.retryTimeout = 'retryTimeout' in config ? config.retryTimeout : 1000;
        if (this.options.headers['Cookie']) {
            this.cookieContainer.addCookies(url, this.options.headers['Cookie']);
        }
        if (config.cookies) {
            this.cookieContainer.addCookies(url, config.cookies);
        }
        if (config.cookiesDefault) {
            this.cookieContainer.addCookies(url, config.cookiesDefault, { extend: true });
        }
        let cookies = this.cookieContainer.getCookies(url);
        if (cookies) {
            this.options.headers['Cookie'] = cookies;
        }
        url = serializeUrl(url, config);
        if (config.agent) {
            this.options.agent = config.agent;
        }
        else {
            if (url.startsWith('http:')) {
                this.options.agent = agents.http;
            }
            if (url.startsWith('https:')) {
                this.options.agent = agents.https;
            }
        }
        if (config.httpsProxy) {
            process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";
            const HttpsProxyAgent = require('https-proxy-agent');
            this.options.agent = new HttpsProxyAgent(config.httpsProxy);
        }
        if (config.body != null && atma_utils_1.is_rawObject(config.body)) {
            Body_1.Body.handleAsRawObject(this.options);
        }
        this.redirectCount = this.options.follow == null ? 10 : this.options.follow;
        this.options.redirect = 'manual';
        this.location = url;
    }
    load() {
        return __awaiter(this, void 0, void 0, function* () {
            let cached = yield this._fromCache();
            if (cached) {
                return cached;
            }
            return yield this._fetch(this.location);
        });
    }
    _fromCache() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                let cached = yield Cache_1.cache.get(this.url, this.config);
                if (cached) {
                    return {
                        status: cached.status,
                        url: cached.url,
                        headers: cached.headers,
                        body: cached.body
                    };
                }
            }
            catch (error) {
                // Not cached
            }
            return null;
        });
    }
    _handleResponse(res) {
        return __awaiter(this, void 0, void 0, function* () {
            let errored = res.status >= 400;
            if (errored && --this.retryCount > 0) {
                switch (res.status) {
                    case 404:
                    case 401:
                    case 403:
                        break;
                    default: {
                        console.log(`Retry ${this.retryCount} for [${this.options.method}] ${this.location} as got ${res.status}`);
                        yield wait(this.retryTimeout);
                        return this._fetch(this.location);
                    }
                }
            }
            let setCookie = res.headers.get('set-cookie');
            if (setCookie) {
                this.cookieContainer.addCookies(this.location, setCookie);
            }
            if (res.status === 301 || res.status === 302) {
                let cookies = this.cookieContainer.getCookies(this.location);
                if (cookies) {
                    this.options.headers['Cookie'] = cookies;
                }
                var location = res.headers.get('location');
                if (!location) {
                    throw new Error('Location not present');
                }
                if (++this.redirectIndex < this.redirectCount) {
                    this.options.method = 'GET';
                    this.options.body = null;
                    if (this.options.headers) {
                        delete this.options.headers['Content-Type'];
                        delete this.options.headers['content-type'];
                        delete this.options.headers['Content-Length'];
                        delete this.options.headers['content-length'];
                    }
                    this.location = location;
                    return this._fetch(location);
                }
            }
            return yield this._handleCompletion(res);
        });
    }
    _handleCompletion(res) {
        return __awaiter(this, void 0, void 0, function* () {
            let errored = res.status >= 400;
            let typeEnum = 'buffer';
            let contentType = res.headers.get('content-type');
            if (contentType && contentType.includes('json')) {
                typeEnum = 'json';
            }
            if (contentType && contentType.includes('text')) {
                typeEnum = 'text';
            }
            let body = null;
            switch (typeEnum) {
                case 'text':
                    body = yield res.textConverted();
                    break;
                case 'json':
                    body = yield res.json();
                    break;
                case 'buffer':
                    let arr = yield res.arrayBuffer();
                    body = Buffer.from(arr);
                    break;
            }
            let resp = {
                status: res.status,
                headers: readAllHeaders(res.headers),
                url: res.url,
                body
            };
            if (errored) {
                let error = new Error(`Request for ${res.url} failed with ${res.status}`);
                error.status = res.status;
                error.body = res.body;
                error.headers = res.headers;
                throw error;
            }
            Cache_1.cache.save(this.location, this.config, resp);
            return resp;
        });
    }
    _fetch(url) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield fetch(url, this.options);
            return this._handleResponse(res);
        });
    }
}
function wait(ms) {
    return new Promise(resolve => {
        setTimeout(() => resolve(), ms);
    });
}
